import awscam
import json
import cv2
import greengrasssdk
import os
from local_display import LocalDisplay
import numpy as np
from time import time
# from matplotlib import pyplot as plt
from mxnet import nd
from gluoncv import model_zoo, data, utils
from gluoncv.data.transforms.pose import detector_to_simple_pose, heatmap_to_coord
from math import atan2, degrees
import datetime
import boto3
import csv

s3 = boto3.client("s3")
S3_BUCKET = '<YOUR_S3_BUCKET_NAME>'
DATA_BUFFER = 60

def lambda_handler(event, context):
    """Empty entry point to the Lambda function invoked from the edge."""
    print("hello world")
    return


def write_to_s3(localfile, bucket, objectname):
    s3.upload_file(localfile, bucket, objectname)


def hip_and_hunch_angle(left_array):
    '''

    :param left_array: pass in the left most coordinates of a person , should be ok, since from side left and right overlap
    :return:
    '''
    # hip to knee angle
    hipX = left_array[-2][0] - left_array[-3][0]
    hipY = left_array[-2][1] - left_array[-3][1]

    # hunch angle = (hip to shoulder ) - (shoulder to ear )
    # (hip to shoulder )
    hunchX1 = left_array[-3][0] - left_array[-6][0]
    hunchY1 = left_array[-3][1] - left_array[-6][1]

    ang1 = degrees(atan2(hunchY1, hunchX1))

    # (shoulder to ear)
    hunchX2 = left_array[-6][0] - left_array[-7][0]
    hunchY2 = left_array[-6][1] - left_array[-7][1]
    ang2 = degrees(atan2(hunchY2, hunchX2))

    return degrees(atan2(hipY, hipX)), abs(ang1 - ang2)


def sitting_and_hunching(left_array):
    hip_ang, hunch_ang = hip_and_hunch_angle(left_array)
    if hip_ang < 25 or hip_ang > 155:
        print("sitting")
        hip = 0
    else:
        print("standing")
        hip = 1
    if hunch_ang < 3:
        print("no hunch")
        hunch = 0
    else:
        hunch = 1
    return hip, hunch


def update_state_json(left_array, client, iot_topic):
    try:
        print("in try section of state update")
        with open("/tmp/state.json", "r+") as f:
            currentJson = json.load(f)
            newJson = currentJson.copy()
            newJson["hip"], newJson["hunch"] = sitting_and_hunching(left_array)
            print("hip : {}, hunch: {}".format(str(newJson["hip"]), str(newJson["hunch"])))
            if newJson["hip"] == 0 and newJson["hip"] == currentJson["hip"]:
                print("in first")
                newJson['hip_time'] = (datetime.datetime.now() - datetime.datetime.strptime(currentJson['hip_start'],
                                                                                            "%Y-%m-%d %H:%M:%S")).seconds / 60.0
                if newJson["hunch"] == 1:
                    newJson['hunch_time'] = currentJson['hunch_time'] + 1

                if newJson['hip_time'] > 30:
                    print("sitting too long for {}, please standup. you were hunching for {}".format(
                        str(newJson['hip_time']), str(newJson['hip_time'] / 60.0)))
                    publishtoSNSTopic(str(newJson['hip_time']), str(newJson['hunch_time']))
            elif newJson['hip'] == 1:
                print("are you standing")
                newJson['hip_time'] = 0
                newJson['hunch_time'] = 0
            else:
                print("in else")
                print("siiting state didn't change")
                newJson['hip_start'] = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

            ## publish to iot topic first
            iotString = json.dumps(newJson)
            cloud_output = {"out": iotString}
            client.publish(topic=iot_topic, payload=json.dumps(cloud_output))

            # update state.json file
            f.seek(0)  # rewind
            json.dump(newJson, f)
            f.truncate()

            if newJson['hip']==0:
                vertical="sitting"
            else:
                vertical="standing"

            if newJson['hunch']==1:
                bend="hunch"
            else:
                bend="no-hunch"
        return vertical, bend
    except Exception as e:
        print("in exception " + str(e))
        with open("/tmp/state.json", "w") as f:
            f.write(json.dumps(
                {"hip": 0, "hip_start": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"), "hip_time": 0,
                 "hunch_time": 0, "hunch": 0, "hunch_start": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}))
        return "",""


def publishtoSNSTopic(SittingTime=None, hunchTime=None):
    sns = boto3.client('sns')

    # Publish a simple message to the specified SNS topic
    # update topic arn
    response = sns.publish(
        TopicArn='arn:aws:sns:us-east-1:973994155064:deeplenspose',
        Message='Alert: You have been sitting for {}, Stand up and stretch, and you have hunched for {}'.format(
            SittingTime, hunchTime),
    )

    print(SittingTime, hunchTime)


def infinite_infer_run():
    """ Run the DeepLens inference loop frame by frame"""
    # Create an IoT client for sending to messages to the cloud.
    client = greengrasssdk.client('iot-data')
    iot_topic = '$aws/things/{}/infer'.format(os.environ['AWS_IOT_THING_NAME'])

    # Create a local display instance that will dump the image bytes to a FIFO
    # file that the image can be rendered locally.
    local_display = LocalDisplay('480p')
    local_display.start()

    MODEL_PATH = '/opt/awscam/artifacts/'

    # Load the model here
    detector = model_zoo.get_model('yolo3_mobilenet1.0_coco', pretrained=True, root=MODEL_PATH)
    # pose_net = model_zoo.get_model('simple_pose_resnet18_v1b')
    # pose_net.load_parameters(MODEL_PATH+'simple_pose_resnet18_v1b-0000.params')
    pose_net = model_zoo.get_model('simple_pose_resnet18_v1b', pretrained=True, root=MODEL_PATH)
    detector.reset_class(["person"], reuse_weights=['person'])
    sec_mark = 1
    coordslist = []
    physicalList=[]
    while True:
        # Get a frame from the video stream
        start = time()
        _, frame = awscam.getLastFrame()
        frame = cv2.resize(frame, (380, 672))
        print('Load frame: {}s'.format(time() - start))
        print(frame.shape)
        start = time()
        x, img = data.transforms.presets.ssd.transform_test(nd.array(frame), short=256)
        print('Shape of pre-processed image:', x.shape)

        print('Load image: {}s'.format(time() - start))

        start = time()
        class_ids, scores, bboxes = detector(x)

        print('Detection: {}s'.format(time() - start))

        pose_input, upscale_bbox = detector_to_simple_pose(img, class_ids, scores, bboxes)

        if pose_input is None:
            print("no person detected")
            continue

        print(pose_input.shape)

        start = time()
        predicted_heatmap = pose_net(pose_input)
        coords, confidence = heatmap_to_coord(predicted_heatmap, upscale_bbox)

        print('Pose Detect: {}s'.format(time() - start))

        # local_display.set_frame_data(out)
        testcoords = coords[0].asnumpy()
        testcoords_left = testcoords[::2]
        # pass the below to the function if deeplens is placed on your right side
        testcoords_right = testcoords[1::2]

        # decide whether to pass left coords or right coords based on how your camera is placed
        print("starting state update")
        vertical , bend= update_state_json(testcoords_left, client, iot_topic)

        dateString = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        if sec_mark == 1:
            try:
                templist = []
                templist2=[]
                templist.append(dateString)
                templist2.append(dateString)
                templist.append(str(testcoords))
                templist2.append(vertical)
                templist2.append(bend)
                x = 0
                coordslist.append(templist)
                physicalList.append(templist2)


                '''
                try:
                    cv2.imwrite("/tmp/temp.png", frame)
                    write_to_s3('/tmp/temp.png', S3_BUCKET,
                                "deeplens-pose/gluoncvpose/fram" + str(
                                    datetime.datetime.now().strftime("%Y-%b-%d-%H-%M-%S")) + ".png")
                except Exception as e:
                    print("couldnt write out to s3 " + str(e))
                    print(type(out))
                '''

            except Exception as e:
                print(e)
        else:
            try:
                sec_mark = int((datetime.datetime.now() - datetime.datetime.strptime(coordslist[-1][0],
                                                                                     "%Y-%m-%d %H:%M:%S")).seconds)
            except Exception as e:
                print("time diff calc error" + str(e))
                print(datetime.datetime.strptime(coordslist[-1][0], "%Y-%m-%d %H:%M:%S"))

        if len(coordslist) > DATA_BUFFER:
            try:
                with open('/tmp/temp.csv', 'w') as f:
                    writer = csv.writer(f)
                    writer.writerows(coordslist)
                coordslist = []
                write_to_s3('/tmp/temp.csv', S3_BUCKET,
                            "deeplens-pose/gluoncvpose/" + datetime.datetime.now().strftime(
                                "%Y-%b-%d-%H-%M-%S") + ".csv")
            except Exception as e:
                print(e)

        if len(physicalList) > DATA_BUFFER:
            try:
                with open('/tmp/temp2.csv', 'w') as f:
                    writer = csv.writer(f)
                    writer.writerows(physicalList)
                physicalList = []
                write_to_s3('/tmp/temp2.csv', S3_BUCKET,
                            "deeplens-pose/gluoncvpose/physicalstate-" + datetime.datetime.now().strftime(
                                "%Y-%b-%d-%H-%M-%S") + ".csv")
            except Exception as e:
                print(e)


infinite_infer_run()